//
// ServletResource.java
//
// Copyright (c) 1997 by Cornell University
// All rights reserved.
//
// Permission is hereby granted, without written agreement and without
// license or royalty fee, to use, copy, modify, and distribute this
// software and its documentation for any purpose, provided that the
// above copyright notice and the following two paragraphs appear in
// all copies of this software
//
// IN NO EVENT SHALL CORNELL UNIVERSITY BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
// ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
// IF CORNELL UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// CORNELL UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
// BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
// ON AN "AS IS" BASIS, AND CORNELL UNIVERSITY HAS NO OBLIGATION TO
// PROVIDE MAINTAINANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//

// (c) COPYRIGHT MIT and INRIA, 1996.
// Please first read the full copyright statement in file COPYRIGHT.html

package w3c.jigsaw.slklet;

import SLK.*;

import java.io.*;
import java.servlet.*;
import java.util.*;

import w3c.util.*;
import w3c.www.http.*;
import w3c.jigsaw.http.*;
import w3c.jigsaw.resources.*;
import w3c.jigsaw.html.*;


public class ServletResource extends FilteredResource implements ServletStub, SLKInterface
{
    private final static boolean debug = false;

    // TODO: need to incorporate the extra path stuff.

    /**
     * Attributes index - The servlet class name.
     */
    protected static int ATTR_SERVLET_CLASS = -1;

    /**
     * Attributes index - The init parameters for that servlet.
     */
    protected static int ATTR_PARAMETERS = -1;

    /**
     * Attributes index - Our parent-inherited servlet context.
     */
    protected static int ATTR_CONTEXT = -1;

    /**
     * Attributes index - The base of a remote servlet.
     */
    protected static int ATTR_SERVLET_BASE = -1;

    /**
     * Attribute index - The SLK keypair.
     */
    protected static int ATTR_SERVLET_KEYPAIR = -1;


    // Stores the servlet.
    private Servlet servlet = null;
    // Stores the wrapper of the servlet.
    protected GenericServletWrapper servletWrapper = null;

    // Stores the class loader corresponding to this servlet.
    protected ServletLoader loader = null;

    // To prevent this servlet to be initialized more than once
    protected boolean initialized = false;

    // For servlet parameters
    protected Hashtable params = null;


    static {
        Attribute a = null;
        Class c = null;
        try {
            c = Class.forName("w3c.jigsaw.slklet.ServletResource");
        }
        catch (Exception e) {
            e.printStackTrace();
        }

        // Registering the attributes
        a = new StringAttribute("servlet-class",
                                null,
                                Attribute.EDITABLE | Attribute.MANDATORY);
        ATTR_SERVLET_CLASS = AttributeRegistry.registerAttribute(c, a);

        a = new PropertiesAttribute("servlet-parameters",
                                    null,
                                    Attribute.EDITABLE);
        ATTR_PARAMETERS = AttributeRegistry.registerAttribute(c, a);

        a = new ObjectAttribute("servlet-context",
                                "w3c.jigsaw.slklet.ServletDirectory",
                                null,
                                Attribute.DONTSAVE);
        ATTR_CONTEXT = AttributeRegistry.registerAttribute(c, a);

        a = new StringAttribute("servlet-base", null,
                                Attribute.EDITABLE | Attribute.MANDATORY);
        ATTR_SERVLET_BASE = AttributeRegistry.registerAttribute(c, a);


        a = new ObjectAttribute("servlet-keypair",
                                "SLK.KeyPair",
                                null,
                                Attribute.MANDATORY | Attribute.DONTSAVE);
        ATTR_SERVLET_KEYPAIR = AttributeRegistry.registerAttribute(c, a);
    }

    /**
     * Launches the servlet.
     * @param c The servlet's class.
     * @return A boolean, <strong>true</strong> if servlet was successfully
     * initialized, <strong>false</strong> otherwise.
     *
     * <p> Makes a new instance of the servlet class and wraps it with one of
     * the three types of wrappers:
     * <ul>
     * <li> ServletWrapperCP - (CallerProtected) single-threaded calls where only
     * Jigsaw has the rights to perform actions on the thread.
     * <li> ServletWrapperCCP - (CallerCalleeProtected) single-threaded calls where
     * neither Jigsaw nor the servlet has the rights to perform any actions on the
     * thread.
     * <li> ServletWrapperTS - (ThreadSwitching) two threads, full protection.
     *
     * <p> The ServletStubWrapperCP protects the servlet's thread during upcalls
     * into Jigsaw (CallerProtected only).
     */

    protected boolean loadServlet(Class c)
    {
        try {
            servlet = (Servlet) c.newInstance();
            servletWrapper = (GenericServletWrapper) new ServletWrapperCP(servlet);
            if (debug) {
                System.out.println("Initializing servlet...");
            }
            try {
                servletWrapper.init(new ServletStubWrapperCP((ServletStub)this));
            }
            catch (NullPointerException e) {
                return (servlet != null);
            }
        }
        catch (Exception e) {
            String msg = ("Failed with exception " + e.getMessage());
            if (debug) {
                e.printStackTrace();
            }
            getServer().errlog(this, msg);
            return false;
        }
        return (servlet != null);
    }

    protected boolean loadServlet()
    {
        if (servlet != null) destroyServlet();
        String cname = getServletClass();
        if (cname == null) {
            getServer().errlog(this, "no servlet class attribute defined.");
            return false;
        }
        else {
            if (debug) {
                System.out.println("Loading servlet " + cname);
            }
            Class c = null;
            try {
                //
                // Load the servlet through the class loader
                // c = Class.forName(cname);
                //
                c = getServletLoader().loadClass(cname, true);

            }
            catch (ClassFormatError e1) {
                String msg = ("Failed.  Class " + cname + " loaded from "+
                              getServletBase() + " has invalid class format");
                if (debug) {
                    e1.printStackTrace();
                }
                getServer().errlog(this, msg);
            }
            catch (ClassNotFoundException e2) {

                String msg = ("Failed.  Class " + cname + " loaded from " +
                              getServletBase() + " is not found");
                if (debug) {
                    e2.printStackTrace();
                }
                getServer().errlog(this, msg);
            }
            return (c != null) ? loadServlet(c) : false;
        }
    }

    /**
     * Dispatch the get request to the servlet.
     * <p> If the servlet cannot be initialized, throw an exception, else pass
     * the request to the servlet's service method.
     * @param request The request to be processed.
     * @exception HTTPException If the wrapped servlet is not initialized.
     */

    public Reply get(Request request)
    throws HTTPException
    {

        if (initialized == false) {
            if (debug) {
                System.out.println("Servlet not initialized.  Initializing...");
            }
            // Try initializing it again.
            initialized = loadServlet();
            if (initialized == false) {
                if (debug) {
                    System.out.println("Initialization failed.");
                }
                Reply reply = request.makeReply(HTTP.INTERNAL_SERVER_ERROR);
                reply.setContent("Servlet not initialized");
                throw new HTTPException(reply);
            }
        }
        Reply reply = createDefaultReply(request, HTTP.OK);
        if (debug) {
            System.out.println("Calling service...");
        }
        try {
            servletWrapper.service(new JigsawServletRequest(request),
                                   new JigsawServletResponse(request, reply));
        }
        catch (Exception e) {
            if (debug) e.printStackTrace();
            reply = request.makeReply(HTTP.INTERNAL_SERVER_ERROR);
            reply.setContent("Failed with servlet exception " + e.toString());
        }
        return reply;
    }


    public synchronized void setValue(int idx, Object value)
    {
        super.setValue(idx, value);
        if ((idx == ATTR_SERVLET_CLASS) && (value != null)) {
            initialized = loadServlet();
        }
    }

    public synchronized void destroyServlet()
    {
        if (servlet != null) {
            servletWrapper.destroy();
            servlet = null;
        }
    }

    /**
     * Lookup on servlets. Need to incorporate the extra path stuff.
     * @param ls The lookup state.
     * @param lr The lookup result.
     * @exception HTTPException If some error occurs.
     */

    public boolean lookup(LookupState ls, LookupResult lr)
    throws HTTPException
    {
        lr.setTarget(this);
        return super.lookup(ls, lr);
    }

    /**
     * Initialize this servlet resource when it is brought from stable
     * storage. After the resource is initialized, it performs the servlet
     * initialization.
     * @param values The default attribute values.
     */
    public void initialize(Object values[])
    {
        super.initialize(values);

        if (getServletClass() != null) {
            initialized = loadServlet();
        }
    }

    //
    // Methods to retrieve protected variables.
    //

    public GenericServletWrapper getServletWrapper()
    {
        return servletWrapper;
    }

    // Fetch the classloader for the servlet.
    protected synchronized ServletLoader getServletLoader()
    {
        if (loader == null) {
            loader = new ServletLoader(this);
        }
        return loader;
    }


    //
    // Returns the servlet class attribute.
    //
    public String getServletClass()
    {
        return getString(ATTR_SERVLET_CLASS, null);
    }

    //
    // Get the remote servlet URL base attribute.
    //
    public String getServletBase()
    {
        return getString(ATTR_SERVLET_BASE, null);
    }


    // ------------------------------------------------------
    // ------------------------------------------------------

    //
    // SERVLETSTUB IMPLEMENTATION
    // Returns the context attribute for the servlet.
    //

    public ServletContext getServletContext ()
    {
        return new ServletContextWrapperCP((ServletContext) getValue(ATTR_CONTEXT, null));
    }

    // SERVLETSTUB IMPLEMENTATION
    // Returns the servlet parameters attribute.
    //
    public ArrayDictionary getServletParameters()
    {
        return (ArrayDictionary) getValue(ATTR_PARAMETERS, null);
    }

    //
    // SERVLETSTUB IMPLEMENTATION
    // Gets an initialization parameter of the servlet
    //
    public synchronized String getInitParameter(String name)
    {
        ArrayDictionary d = getServletParameters();
        String v = (d != null) ? (String) d.get(name) : null;
        return v;
    }

    //
    // SERVLETSTUB IMPLEMENTATION
    // Returns a hashtable of initialization parameters for the servlet
    //
    public Hashtable getInitParameters()
    {
        ArrayDictionary d = getServletParameters();
        if (d != null) {
            Enumeration keys = d.keys();
            params = new Hashtable(d.size());
            while (keys.hasMoreElements()) {
                Object key = keys.nextElement();
                params.put(key, d.get(key));
            }
        }
        return params;
    }

    //
    // SLKInterface Implementation
    //

    /**
     * Retrieve the SLK keypair for this servlet.
     * @return KeyPair The SLK key pair.
     */
    public KeyPair getKeyPair()
    {
        return (KeyPair) getValue(ATTR_SERVLET_KEYPAIR, null);
    }
}
